#ifndef LAB_GL_UTILS_H
#define LAB_GL_UTILS_H

/**
 * This file contains utility functions to be used for the labs in computer graphics at chalmers,
 * they are not covered by any particular license...
 */

#include "float4x4.h"
#include "float3.h"
#include <string>
#include <cassert>


// Sometimes it exists, sometimes not...
#ifndef M_PI
  #define M_PI	3.14159265358979323846f
#endif 

chag::float4x4 perspectiveMatrix(float fov, float aspectRatio, float n, float f); /* field of view, aspect ratio, near clipping plane, far clipping plane */
chag::float4x4 lookAt(const chag::float3 &eyePosition, const chag::float3 &lookAt, const chag::float3 &desiredUp);
GLuint loadCubeMap(const char* facePosX, const char* faceNegX, const char* facePosY, const char* faceNegY, const char* facePosZ, const char* faceNegZ);

/**
 * This macro checks for gl errors using glGetError, it is useful to sprinkle it around the code base, especially
 * when unsure about the correct usage, for example after each call to open gl.
 * When the debugger is atached it will cause a break on the offending line, and also print out the file 
 * and line in a MSVC compatible format on the debug output and console.
 *
 * Note: the macro _cannot_ be used between glBegin/glEnd brackets, as stated by the openGL standard.
 * Note2: be aware that the macro will report any previous errors, since the last call to glGetError.
 *
 * example usage: glClear(GL_COLOR_BUFFER_BIT); CHECK_GL_ERROR(); // this will check for errors in this (and any previous statements)
 */

#if !defined(_WIN32)
#	define __debugbreak() assert(false)
#endif
#define CHECK_GL_ERROR() { checkGLError(__FILE__, __LINE__) && (__debugbreak(), 1); }
/**
 * Internal function used by macro CHECK_GL_ERROR, use that instead.
 */
bool checkGLError(const char *file, int line);
/** 
 * Error reporting function
 */
void fatal_error( std::string errorString, std::string title = std::string() );



/**
 * Helper function used to create a shader object from text in a file
 */
const char *textFileRead(const char *fn);

/**
 * Helper function used to get log info (such as errors) about a shader object or shader program
 */
std::string GetShaderInfoLog(GLuint obj);

/**
 * Loads and compiles a fragment and vertex shader. Then creates a shader program
 * and attaches the shaders. Does NOT link the program, this is done with  linkShaderProgram()
 * The reason for this is that before linking we need to bind attribute locations, using
 * glBindAttribLocation and fragment data lications, using glBindFragDataLocation.
 */
GLuint loadShaderProgram(const std::string &vertexShader, const std::string &fragmentShader);
/**
 * Call to link a shader program prevoiusly loaded using loadShaderProgram.
 */
void linkShaderProgram(GLuint shaderProgram);

/**
 * Creates a GL buffer and uploads the given data to it.
 * returns the handle of the GL buffer.
 */
GLuint createAddAttribBuffer(GLuint vertexArrayObject, const void *data, const size_t dataSize, GLuint attributeIndex, GLsizei attributeSize, GLenum type, GLenum bufferUsage = GL_STATIC_DRAW);


/**
 * Helper to set uniform variables in shaders, labeled SLOW because they find the location from string each time.
 * In OpenGL (and similarly in other APIs) it is much more efficient (in terms of CPU time) to keep the uniform
 * location, and use that. Or even better, use uniform buffers!
 * However, in the simple tutorial samples, performance is not an issue.
 * Overloaded to set many types.
 */
void setUniformSlow(GLuint shaderProgram, const char *name, const chag::float4x4 &matrix);
void setUniformSlow(GLuint shaderProgram, const char *name, const float value);
void setUniformSlow(GLuint shaderProgram, const char *name, const GLint value);
void setUniformSlow(GLuint shaderProgram, const char *name, const chag::float3 &value);

/**
 * Code that draws a sphere where the light is and a stippled line to the 
 * origo. Note that this method uses functions that have been deprecated in
 * OpenGL 3 and so would not necessarily work on all future implementations. 
 * 
 * That said, for drawing simple debug information such as this, OpenGL 2
 * can be pretty handy.
 */
void debugDrawLight(const chag::float4x4 &viewMatrix, 
					const chag::float4x4 &projectionMatrix, 
					const chag::float3 &worldSpaceLightPos);

#endif // LAB_GL_UTILS_H
